# -*- coding: utf-8 -*-
#
# Author: Chris Oliver (excid3@gmail.com)
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA

import commands, os, os.path, platform, gzip, lib
from lib import plugins, consts, log

PLUGIN_NAME = 'Debian'
PLUGIN_TYPE = 'OS'
FILE = 'debian.settings'

class Debian(plugins.pluginBase):        
    def IsOS(self): # Returns bool if current OS is a match
        if platform.dist()[0] == 'debian': return True
        return False

    def createProject(self, name): # Creates project files
        dirNew = os.path.join(consts.dirProjects, name) # Create in projects directory
        
        try:
            # Set information
            info = self.__setInfo()
            self.__grabFiles(name, dirNew)

            outfile = open(os.path.join(dirNew, FILE), 'wb')
            outfile.write(_('Computer Name: ') + info[0] + '\n'
                          + _('OS Name: ') + info[1] + '\n'
                          + _('OS Version: ') + info[2] + '\n'
                          + _('Architecture: ') + info[3] + '\n'
                          + _('Kernel: ') + info[4] + '\n')
            outfile.close()
            return True
        except:
            log.error('Problem writing to project settings file')
        return False

    def loadProject(self, directory):
        infile = open(os.path.join(directory, FILE), 'rb')    
        data = infile.read()
        infile.close()
        data = data.split()
        return [data[2], data[5], data[8], data[10], data[12]]

    def loadLocalPackageList(self, dir, arch):
        listDir = os.path.join(dir, 'lists')
        debs = self.__parseSources(dir)
        filenames = self.__filesFromDebs(debs, arch)
        installed = self.__getInstalled(dir)

        packages = []
        count = 0

        for item in filenames:
            try:
                data = open(os.path.join(listDir,item), 'rb')
                packages.append(self.__readPackages(data, installed, 'http://' + item.split('_dists')[0].replace('_','/')))
                data.close()
                count += 1
            except Exception, exc:
                log.error(_("Unable to open file: ") + str(exc))
 
        # Count packages
        total=0 
        for item in packages: total += len(item)
 
        # If no packages are found, see if we have package lists at least
        if total == 0:
            log.error(_('No packages found'))
            log.info(_('Loading any package lists possible'))
            for item in os.listdir(listDir):
                if item.endswith('Packages'):
                    try:
                        data = open(os.path.join(listDir, item), 'rb')
                        packages.append(self.__readPackages(data, installed, 'http://' + item.split('_dists')[0].replace('_','/')))
                        data.close()
                        count+=1
                    except Exception, exc:
                        log.error(_("Unable to open file: ") + str(exc))
 
        return packages

    def loadInternetPackageList(self, dir, arch):
        listDir = os.path.join(dir, "lists")
        debs = self.__parseSources(dir)
        filenames = self.__filesFromDebs(debs, arch, listDir)
        tempnames = self.__tempFilesFromDebs(debs, arch, listDir)
        urls = self.__urlsFromDebs(debs, arch)
        return zip(urls, tempnames, filenames) # Returns urls, gzip file locations, and extracted locations

    def getDependencies(self, allPackages, packageName):
        # Recusive function, takes package name, returns package information for each dependency
        urls = []
#        return urls
        if not allPackages.has_key(packageName): return {}
        values = allPackages.get(packageName) 
        dependencies = values[5].split(", ")
        filename = values[6].split("/") # Get the parts for the package filename
        urls.append((values[6], os.path.join(consts.dirPackages, filename[len(filename)-1])))
        values[2] = values[3]
        allPackages[packageName] = values

        for item in dependencies:
            data = item.split()
            if data == []: break # No dependencies :D

            # Get dependencies for this package too if it isnt already installed
            if allPackages.has_key(data[0]) and allPackages.get(data[0])[2] == "":
                urls += self.getDependencies(allPackages, data[0])

        return urls

    def __readPackages(self, infile, installed, mainUrl=''):
        packages = {}
        # up-to-date, name, installed ver, latest ver, descrip, depends, filename, size(int)
        current = ['','','','','','','','']

        for line in infile:
            if line.startswith("Package:"):     current[1] = line[9:-1]
            if line.startswith("Version:"):     current[3] = line[9:-1]
            if line.startswith("Description:"): current[4] = lib.utf(line[13:-1])
            if line.startswith("Depends:"):     current[5] = line[9:-1]
            if line.startswith("Filename:"):    current[6] = lib.joinUrl(mainUrl, line[10:-1])
            if line.startswith("Size:"):        current[7] = int(line[6:-1])

            if line.startswith("\n") and current[1] != '':
                status = self.__checkPackage(current, installed, packages)
                if status != '': packages.update({current[1] : status})
                name = ''
                current = ['','','','','','','','']
                
        return packages

    def __checkPackage(self, package, installed, packages):
        # Check to see if package already exists
#        print packages, installed, package
#        for item in packages:
#            if item[0] > package[2]: break # Already passed where the package would have been
#            if package[2] == item[0]: # Does exist
#                # Check to see if package is newer than item
#                if package[4] > item[1]: # Current package is newer than the one in the list
#                    package[3] = package[4]
#                    package[1] = gtk.STOCK_APPLY
#                    return package
#                if package[4] <= item[1]:
#                    return ''

        # Doesnt already exist in the list
        for item in installed:
            if package[1] == item[0]:

                #Compare version numbers, set icon accordingly
                if item[1] == package[3]:  package[0] = 0 # up-to-date
                elif item[1] < package[3]: package[0] = 1 # update
                elif item[1] > package[3]: package[0] = 2 # error
                package[2] = item[1]
                return package
        return package

    def __setInfo(self): # Sets the Debian project info
        uname = platform.uname()
        comp_name = uname[1]
        os_arch = platform.architecture()[0]
        os_kern = uname[2]
        try:
            status = commands.getstatusoutput('cat /etc/issue.net')
            if status[0] != 0: log.error(_('Problem retrieving Debian version'))
            temp = status[1].split()
            os_name = temp[0]
            os_ver = temp[1]
        except:
            os_name = os_ver = _('Unknown')
        return [comp_name, os_name, os_ver, os_arch, os_kern] 

    def __grabFiles(self, name, folder): # Grabs files necessary for a Debian project
        import shutil
        dirLists = os.path.join(folder, 'lists')
        dirSources = os.path.join(folder, 'sources')

        try:    shutil.copytree('/etc/apt/', dirSources)
        except: pass # Will always raise errors because some files are only readable by root
        try:    shutil.copytree('/var/lib/apt/lists/', dirLists)
        except: pass
        shutil.copyfile('/var/lib/dpkg/status', os.path.join(dirLists,'status'))

    def __parseSources(self, dir): # Returns list of all deb entries for project
        sourcesDir = os.path.join(dir, 'sources')
        sourcesdDir = os.path.join(sourcesDir, 'sources.list.d')
        sources = self.__parseFile(os.path.join(sourcesDir, 'sources.list'))
        if os.path.exists(sourcesdDir):
            sourcesd = os.listdir(sourcesdDir)
            for item in sourcesd:
                if item.endswith('.list'): sources += self.__parseFile(os.path.join(sourcesdDir, item))
        return sources        

    def __parseFile(self, location): # Returns a sepcific file's deb entries
        found = []
        infile = open(location)
        for line in infile:
            if line.startswith('deb http://'): #TODO: Add support for more protocols
                if line.endswith('\n'): found.append(line[:-1])
                else:                   found.append(line)
        return found

    def __filesFromDebs(self, deblist, arch, dir=''): # Generates a list of files from deb entries
        local = []
        for item in deblist:
            data = item.split()
            try:
                dtype = data[0]
                url = data[1]
                dist = data[2]
                if len(data) == 3: pass #FIXME: Special case, append only the sections to the end
                for section in data[3:]:
                    if section.find('#') != -1: break # If a comment is encountered skip the line
                    main = lib.joinUrl(lib.joinUrl(lib.joinUrl(url, 'dists'), dist), section)
                    main = self.__appendArch(arch, main)
                    main = main[7:-3].replace('/','_')
                    local.append(os.path.join(dir, main)) # Strips unnecessary characters and appends to list
            except: pass # Unable to parse deb entry
        return local
        
    def __tempFilesFromDebs(self, deblist, arch, dir=''): # Generates a list of files from deb entries
        local = []
        for item in deblist:
            data = item.split()
            try:
                dtype = data[0]
                url = data[1]
                dist = data[2]
                if len(data) == 3: pass #FIXME: Special case, append only the sections to the end
                for section in data[3:]:
                    if section.find('#') != -1: break # If a comment is encountered skip the line
                    main = lib.joinUrl(lib.joinUrl(lib.joinUrl(url, 'dists'), dist), section)
                    main = self.__appendArch(arch, main)
                    main = main[7:].replace('/','_')
                    local.append(os.path.join(dir, main)) # Strips unnecessary characters and appends to list
            except: pass # Unable to parse deb entry
        return local
        
    def __urlsFromDebs(self, deblist, arch):
        urls = []
        for item in deblist:
            data = item.split()
            try:
                dtype = data[0]
                url = data[1]
                dist = data[2]
                if len(data) == 3: pass
                for section in data[3:]:
                    if section.find('#') != -1: break
                    main = lib.joinUrl(lib.joinUrl(lib.joinUrl(url, 'dists'), dist), section)
                    main = self.__appendArch(arch, main)
                    urls.append(main)
            except: pass
        return urls
        

    def __appendArch(self, arch, location): # Appends an architecture to a location
        if arch == '32bit': return lib.joinUrl(location, 'binary-i386/Packages.gz')
        elif arch == '64bit': return lib.joinUrl(location, 'binary-amd64/Packages.gz')
        return location

    def __getInstalled(self, dir): # Gets a list of installed packages
        status = open(os.path.join(os.path.join(dir, 'lists'),'status'), 'rb')

        installed = []
        current = ['', '', ''] # name, version, status
        for line in status:
            if line.startswith('Package:'): current[0] = line[9:-1]
            if line.startswith('Version:'): current[1] = line[9:-1]
            if line.startswith('Status:'):  current[2] = line[8:-1]
            if line.startswith('\n') and current[2] == 'install ok installed':
                installed.append(current)
                current = ['', '', '']

        status.close()
        return installed
